import pkgutil
import inspect
import logging
import datetime
import collections
from importlib import import_module
from .exceptions import Misconfigured
import mixins
import config
import es


_elasticsearch_indices = collections.defaultdict(lambda: [])


logger = logging.getLogger(__name__)


def find_search_types(module, base=mixins.ElasticsearchIndexMixin):
    """Returns a list of automatically detected search type
    classes. Useful for populating ELASTICSEARCH_TYPE_CLASSES.

    :param module: The module to search.
    :type module: A Python module.
    :param base: A class that your search type classes inherit from.
    :type base: A Python class.
    :returns:  A list of type classes that inherit from base.
    """
    retrieved_types = []

    for importer, modname, ispkg in pkgutil.iter_modules(module.__path__):
        module_to_load = '.'.join([module.__name__, modname])
        obj = import_module(module_to_load)
        if ispkg:
            retrieved_types.extend(find_search_types(obj, base))
        else:
            module_klasses = inspect.getmembers(
                obj,
                lambda x: inspect.isclass(x) and
                issubclass(x, base)
            )

            for name, klass in module_klasses:
                if klass is not base and klass not in retrieved_types:
                    retrieved_types.append(klass)

    return retrieved_types


def get_indices(indices=None):
    """Returns indices and their search type classes

    :param indices: Filter returned indices by this list.
    :type indices: list.
    :returns:  A dictionary containing indices as keys and
               a list of search classes as values.
    :raises: Misconfigured
    """
    if not _elasticsearch_indices:
        type_classes = getattr(
            config.settings,
            'ELASTICSEARCH_TYPE_CLASSES',
            ()
        )

        if not type_classes:
            raise Misconfigured(
                'Missing ELASTICSEARCH_TYPE_CLASSES in settings.'
            )

        for type_class in type_classes:
            if inspect.isclass(type_class):
                klass = type_class
            else:
                package_name, klass_name = type_class.rsplit('.', 1)
                try:
                    package = import_module(package_name)
                    klass = getattr(package, klass_name)
                except (ImportError, AttributeError):
                    logger.error('Unable to import `{}`.'.format(type_class))
                    continue

            _elasticsearch_indices[klass.get_index_name()].append(klass)

    result = {}
    if not indices:
        result = _elasticsearch_indices.copy()
    else:
        for k, v in _elasticsearch_indices.items():
            if k in indices:
                result[k] = v

    if not result:
        raise Misconfigured(
            'ELASTICSEARCH_TYPE_CLASSES configured, but none could be loaded.'
        )

    return result


def get_indices_from_aliases(search_aliases=None):
    """Returns indices associated with aliases

    :param search_aliases: Search aliases to get indices for.
    :type search_aliases: list.
    :returns:  A list of indices
    """
    if not search_aliases:
        search_aliases = []

    indices = []

    for alias in search_aliases:
        indices.extend([
            index
            for index, aliases in es.get_aliases().items()
            if alias in aliases['aliases'].keys()
        ])

    return indices


def get_alias_names(aliases=None):
    """Return unique names from aliases

    :param aliases: A list alias/index tuples to get names of.
    :type search_aliases: list
    :returns:  A unique list of alias names
    """
    return list(set([alias[0] for alias in aliases or []]))


def generate_aliases_for_removal():
    current_aliases = es.get_aliases()
    aliases_for_removal = collections.defaultdict(lambda: [])
    for item, tmp in current_aliases.items():
        for iname in list(tmp.get('aliases', {}).keys()):
            aliases_for_removal[iname].append(item)

    return aliases_for_removal


def generate_actions(indices=None, aliases_for_removal=None):
    actions = []
    indices = indices or []
    aliases_for_removal = aliases_for_removal or collections.defaultdict(lambda: [])

    for index_alias, index_name in indices:
        for item in aliases_for_removal[index_alias]:
            actions.append({
                'remove': {
                    'index': item,
                    'alias': index_alias
                }
            })
        actions.append({
            'add': {
                'index': index_name,
                'alias': index_alias
            }
        })

    return actions


def create_aliases(indices=None):
    indices = indices or []

    aliases_for_removal = generate_aliases_for_removal()
    actions = generate_actions(indices, aliases_for_removal)
    es.update_aliases({'actions': actions})


def create_indices(indices=None, set_aliases=True):
    result = []
    aliases = []
    indices = indices or []

    now = datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
    for index_alias, type_classes in get_indices(indices).items():
        index_settings = recursive_dict_update(
            getattr(
                config.settings,
                'ELASTICSEARCH_DEFAULT_INDEX_SETTINGS',
                {}
            ),
            getattr(
                config.settings,
                'ELASTICSEARCH_CUSTOM_INDEX_SETTINGS',
                {}
            ).get(index_alias, {})
        )

        index_name = '{0}-{1}'.format(index_alias, now)

        aliases.append((index_alias, index_name))

        type_mappings = {}
        for type_class in type_classes:
            tmp = type_class.get_type_mapping()
            if tmp:
                type_mappings[type_class.get_type_name()] = tmp

            result.append((
                type_class,
                index_alias,
                index_name
            ))

        # if we got any type mappings, put them in the index settings
        if type_mappings:
            index_settings['mappings'] = type_mappings

        es.create_index(index_name, index_settings)

    if set_aliases:
        create_aliases(aliases)

    return result, aliases


def delete_indices(indices=None):
    es.delete_indices(indices or [])


def rebuild_indices(indices=None, set_aliases=True):
    indices = indices or []
    created_indices, aliases = create_indices(indices, False)

    current_index_name = None

    def change_index():
        if current_index_name:
            es.refresh_index(current_index_name)

    for type_class, index_alias, index_name in created_indices:
        if index_name != current_index_name:
            change_index()
            current_index_name = index_name

        try:
            type_class.bulk_index(index_name)
        except NotImplementedError:
            logger.error(
                'bulk_index not implemented on {} - {}.'.format(
                    type_class.get_index_name(),
                    type_class.__name__
                )
            )
            continue
    else:
        change_index()

    if set_aliases:
        alias_names = get_alias_names(aliases)
        existing_aliased_indices = get_indices_from_aliases(alias_names)

        create_aliases(aliases)

        new_aliased_indices = get_indices_from_aliases(alias_names)

        for index in existing_aliased_indices:
            # Ensure that there are new aliased indices, and that our old
            # index is not somehow in them.
            if new_aliased_indices and index not in new_aliased_indices:
                if config.settings.ELASTICSEARCH_DELETE_OLD_INDICES:
                    delete_indices(get_indices(index))

    return created_indices, aliases


def recursive_dict_update(d, u):
    for k, v in u.items():
        if isinstance(v, collections.Mapping):
            r = recursive_dict_update(d.get(k, {}), v)
            d[k] = r
        else:
            d[k] = u[k]
    return d



"""
An example configuration:

    ELASTICSEARCH_SETTINGS = {
        'ELASTICSEARCH_DELETE_OLD_INDICES': True,
        'ELASTICSEARCH_INDEX_NAME': env(
            'ELASTICSEARCH_INDEX_NAME',
            'example'
        ),
        'ELASTICSEARCH_TYPE_CLASSES': [
            'api.search.VideoSearch',
            'api.search.PlaylistSearch'
        ],
        'ELASTICSEARCH_CONNECTION_PARAMS': {
            'hosts': env(
                'ELASTICSEARCH_SERVER',
                'http://127.0.0.1:9200'
            ),
        },
    }
"""

import importlib
import inspect


class SettingsObject(object):
    def __init__(self, *args, **kwargs):
        for dictionary in args:
            for key in dictionary:
                setattr(self, key, dictionary[key])
        for key in kwargs:
            setattr(self, key, kwargs[key])


def configure(new_settings):
    global settings
    settings = SettingsObject(new_settings)


settings = SettingsObject()
_search_types = None


def get_search_types():
    """Creates and returns a dictionary of search type classes
    with the name/class as k/v pairs.

    :returns:  A dictionary of search type classes.
    """
    if not _search_types:
        global _search_types
        _search_types = {
            search_type.get_type_name(): search_type
            for search_type in [
                klass if inspect.isclass(klass) else getattr(
                    importlib.import_module(module),
                    klass
                )
                for module, klass in [
                    (None, st) if inspect.isclass(st) else st.rsplit('.', 1)
                    for st in getattr(
                        settings,
                        'ELASTICSEARCH_TYPE_CLASSES',
                        []
                    )
                ]
            ]
        }

    return _search_types



"""
This contains all external calls to the Elasticsearch client
in a handy single location
"""

from elasticsearch import Elasticsearch
import config


def get_aliases():
    es = Elasticsearch(**config.settings.ELASTICSEARCH_CONNECTION_PARAMS)

    return es.indices.get_aliases()


def get_settings(index_name):
    es = Elasticsearch(**config.settings.ELASTICSEARCH_CONNECTION_PARAMS)

    return es.indices.get_settings(index_name)


def update_aliases(actions):
    es = Elasticsearch(**config.settings.ELASTICSEARCH_CONNECTION_PARAMS)

    es.indices.update_aliases(actions)


def create_index(index_name, index_settings):
    es = Elasticsearch(**config.settings.ELASTICSEARCH_CONNECTION_PARAMS)

    es.indices.create(index_name, index_settings)


def delete_indices(indices=None):
    es = Elasticsearch(**config.settings.ELASTICSEARCH_CONNECTION_PARAMS)

    if not indices:
        indices = []

    for index in indices:
        if es.indices.exists(index):
            es.indices.delete(index)


def refresh_index(index_name):
    es = Elasticsearch(**config.settings.ELASTICSEARCH_CONNECTION_PARAMS)

    es.indices.refresh(index_name)



from unittest import TestCase
from printen.mixins import ElasticsearchIndexMixin
import printen.config


class ConfigTests(TestCase):
    def tearDown(self):
        # Ensure we have a clean settings object
        # between runs
        printen.config.settings = printen.config.SettingsObject()

    def test_settings_setup(self):
        self.assertTrue(
            isinstance(printen.config.settings, printen.config.SettingsObject)
        )

    def test_configure_settings(self):
        with self.assertRaises(AttributeError):
            printen.config.settings.foo
        printen.config.configure({'foo': 'bar'})
        self.assertEqual(printen.config.settings.foo, 'bar')

    def test_get_search_types(self):
        class SearchTypeClass(ElasticsearchIndexMixin):
            pass

        printen.config.configure({
            'ELASTICSEARCH_TYPE_CLASSES': [
                SearchTypeClass,
            ]
        })

        self.assertIs(
            printen.config._search_types,
            None
        )

        search_types = printen.config.get_search_types()

        self.assertEqual(
            search_types,
            {
                'SearchTypeClass': SearchTypeClass
            }
        )

        self.assertEqual(
            printen.config._search_types,
            search_types
        )



import gc
from elasticsearch import Elasticsearch, TransportError
from .exceptions import MissingObjectError
import config


class ElasticsearchIndexMixin(object):
    """
    This class can be used with a standalone class or
    directly on your model. There are several methods
    that you are required to subclass in order to
    use this.
    """
    @classmethod
    def get_es(cls):
        """Returns the configured reference to the Elasticsearch
        instance.

        :returns:  The elasticsearch instance
        :raises: Misconfigured
        """
        if not hasattr(cls, '_es'):
            cls._es = Elasticsearch(**cls.get_es_connection_settings())
        return cls._es

    @classmethod
    def get_es_connection_settings(cls):
        """Returns the connection parameters for Elasticsearch

        :returns:  A dictionary containing key/value pairs of
                    connection parameters
        :raises: AttributeError
        """
        return config.settings.ELASTICSEARCH_CONNECTION_PARAMS

    @classmethod
    def get_index_name(cls):
        """Allows indexing code to know the name of the index to use.
        This defaults to `index` but can be overridden by setting
        the `ELASTICSEARCH_INDEX_NAME` in your configuration, or
        overriding this method on a per-model basis.

        :returns:  A string containing the index name.
        """
        return getattr(config.settings, 'ELASTICSEARCH_INDEX_NAME', 'index')

    @classmethod
    def get_type_name(cls):
        """This provides Elasticsearch with a name with which to
        do type mappings for your model. Defaults to
        cls.__class__.__name__.

        :returns:  A string naming the type of thing being indexed.
        """
        return cls.__name__

    @classmethod
    def get_document_id(cls, obj):
        """Returns the document id, which usually corresponds to
        the objects primary key in the database.

        You *must* override this method.

        :param obj: The object to derive an id from.
        :type obj: A python object which you are indexing.
        :returns:  An integer representing the document's id.
        """
        raise NotImplementedError

    @classmethod
    def get_document(cls, obj):
        """Returns the 'document' to be indexed, which is a dictionary
        representation of the object passed in, including the relevant
        fields you wish to be in the search engine.

        You *must* override this method.

        :param obj: The object to generate a document for.
        :type obj: A python object which you are indexing.
        :returns:  A dictionary with key/value pairs of fields and values.
        """
        raise NotImplementedError

    @classmethod
    def get_request_params(cls, obj):
        """Overridable per-model request parameters for
        Elasticsearch.

        :param obj: The object involved in the request.
        :type obj: A python object which you are indexing.
        :returns:  Request parameters dictionary passed to Elasticsearch's
                   index operation.
        """
        return {}

    @classmethod
    def get_type_mapping(cls):
        """Overridable per-model type mapping for
        Elasticsearch.

        :returns:  Type mapping for this type used when creating the index.
        """
        return {}

    @classmethod
    def get_queryset(cls):
        """Get a queryset for the model.

        You *must* implement this yourself.

        :returns:  An object with an interface that your queryset_iterator
                   method knows how to iterate over.
        """
        raise NotImplementedError

    @classmethod
    def queryset_iterator(cls, queryset):
        """Iterate (using a generator) over a given queryset.

        You *must* implement this yourself.

        :param queryset: The queryset to iterate over.
        :type queryset: A type which provides an interface this
                        method knows how to iterate over.
        :yields:        Instances of the object to index.
        """
        raise NotImplementedError

    @classmethod
    def get_bulk_index_batch_size(cls):
        """Get the number of items to limit bulk indexing to
        at a time (i.e. per call to the Elasticsearch server).

        Defaults to 100. Set `ELASTICSEARCH_BULK_INDEX_BATCH_SIZE`
        to change this value, or override this method.

        :returns:  An integer.
        """
        return getattr(
            config.settings,
            'ELASTICSEARCH_BULK_INDEX_BATCH_SIZE',
            100
        )

    @classmethod
    def should_index(cls, obj):
        """Determines if an obj should be indexed.

        Defaults to True

        :param obj: The object to generate a document for.
        :type obj: A python object which you are indexing.
        :returns:  A Boolean.
        """
        return True

    @classmethod
    def bulk_index(cls, index_name='', queryset=None):
        """Initiate a bulk index of all items returned
        by the queryset.

        :param index_name: The index to update. By default this
                           is the value returned by :func:`get_index_name`
        :param queryset: The queryset to iterate over.
        """
        es = cls.get_es()

        tmp = []

        if queryset is None:
            queryset = cls.get_queryset()

        for obj in cls.queryset_iterator(queryset):
            delete = not cls.should_index(obj)

            data = {
                '_index': index_name or cls.get_index_name(),
                '_type': cls.get_type_name(),
                '_id': cls.get_document_id(obj)
            }
            data.update(cls.get_request_params(obj))
            data = {'delete' if delete else 'index': data}
            tmp.append(data)

            if not delete:
                tmp.append(cls.get_document(obj))

            if not len(tmp) % cls.get_bulk_index_batch_size():
                es.bulk(tmp)
                tmp = []

        if tmp:
            es.bulk(tmp)

    @classmethod
    def index_add(cls, obj, index_name=None):
        """Adds a single object to the index.

        :param obj: The object to index.
        :type obj: A python object which you are indexing.
        :param index_name: The index to update. By default this
                           is the value returned by :func:`get_index_name`
        :returns:  A Boolean indicating if the object was added.
        """
        if obj and cls.should_index(obj):
            cls.get_es().index(
                index_name or cls.get_index_name(),
                cls.get_type_name(),
                cls.get_document(obj),
                cls.get_document_id(obj),
                **cls.get_request_params(obj)
            )
            return True
        return False

    @classmethod
    def index_delete(cls, obj, index_name=None):
        """Removes a single object to the index.

        :param obj: The object to index.
        :type obj: A python object which you are indexing.
        :param index_name: The index to update. By default this
                           is the value returned by :func:`get_index_name`
        :returns:  A Boolean indicating if the object was removed.
        """
        if obj:
            try:
                cls.get_es().delete(
                    index_name or cls.get_index_name(),
                    cls.get_type_name(),
                    cls.get_document_id(obj),
                    **cls.get_request_params(obj)
                )
            except TransportError as e:
                if e.status_code != 404:
                    raise
            return True
        return False

    @classmethod
    def index_add_or_delete(cls, obj, index_name=None):
        """Adds or removes an object depending on the return value
        of :func:`should_index`.

        :param obj: The object to index.
        :type obj: A python object which you are indexing.
        :param index_name: The index to update. By default this
                           is the value returned by :func:`get_index_name`
        :returns:  A Boolean indicating if the object was added/removed.
        """
        if obj:
            if cls.should_index(obj):
                return cls.index_add(obj, index_name)
            else:
                return cls.index_delete(obj, index_name)
        return False

    @classmethod
    def save_handler(cls, sender, instance):
        """A target for wiring to a save signal for your
        model.

        :param sender: The class which initated the save.
        :param instance: The instance being saved.
        """
        cls.index_add_or_delete(instance)

    @classmethod
    def delete_handler(cls, sender, instance, **kwargs):
        """A target for wiring to a delete signal for your
        model.

        :param sender: The class which initated the delete.
        :param instance: The instance being delete.
        """
        cls.index_delete(instance)


class PeeweeElasticsearchIndexMixin(ElasticsearchIndexMixin):
    """
    A class that inherits from ElasticsearchIndexMixin,
    overriding the relevant methods with Peewee related
    defaults.
    """
    @classmethod
    def get_type_name(cls):
        """Provides a sane default for Peewee models. Override
        to modify behavior.

        :returns: `cls.model._meta.name`
        """
        return cls.model._meta.name

    @classmethod
    def get_document_id(cls, obj):
        """Provides a sane default for Peewee models. Override
        to modify behavior.

        :param obj: The object to index.
        :type obj: A subclass of peewee.Model
        :returns: `getattr(obj, obj._meta.primary_key.name)` which is
                  an integer document id
        :raises: MissingObjectError
        """
        try:
            doc_id = getattr(obj, obj._meta.primary_key.name, None)
        except AttributeError:
            doc_id = None

        if not doc_id:
            raise MissingObjectError

        return doc_id

    @classmethod
    def get_queryset(cls):
        """Provides a sane default for Peewee models. Override
        to modify behavior.

        :returns: `cls.model.select()` - a Peewee queryset
        """
        return cls.model.select()

    @classmethod
    def queryset_iterator(cls, queryset):
        """Provides a sane default for Peewee models. Override
        to modify behavior.

        Executes the given queryset and iterates over the results,
        yielding them one at a time to the caller.

        :param queryset: A Peewee queryset to iterate over.
        :type queryset: A `Peewee queryset <http://docs.peewee-orm.com/en/latest/peewee/api.html#SelectQuery>`_.
        :returns: `cls.model.select()` - a Peewee queryset
        """
        for result in queryset.execute().iterator():
            yield result


class DjangoElasticsearchIndexMixin(ElasticsearchIndexMixin):
    """
    A class that inherits from ElasticsearchIndexMixin,
    overriding the relevant methods with Django related
    defaults.
    """
    @classmethod
    def get_type_name(cls):
        """Provides a sane default for Django models. Override
        to modify behavior.

        :returns: String with a dot separated app_label and model_name.
        """
        return '{}.{}'.format(cls._meta.app_label, cls._meta.model_name)

    @classmethod
    def get_document_id(cls, obj):
        """Provides a sane default for Django models. Override
        to modify behavior.

        :param obj: The object to index.
        :type obj: A subclass of `django.db.models.Model <https://docs.djangoproject.com/en/1.9/ref/models/instances/#django.db.models.Model>`_.
        :returns: `obj.pk`, an integer.
        :raises: MissingObjectError
        """
        if not obj:
            raise MissingObjectError

        try:
            doc_id = obj.pk
        except AttributeError:
            doc_id = None

        if not doc_id:
            raise MissingObjectError

        return obj.pk

    @classmethod
    def get_queryset(cls):
        """Provides a sane default for Django models. Override
        to modify behavior.

        :returns: A `Django QuerySet <https://docs.djangoproject.com/en/1.9/ref/models/querysets/>`_..
        """
        return cls.objects.all()

    def queryset_iterator(self, queryset, chunksize=1000):
        """Provides a sane default for Django models. Override
        to modify behavior.

        :param queryset: A Django QuerySet to iterate over.
        :type queryset: A `Django QuerySet <https://docs.djangoproject.com/en/1.9/ref/models/querysets/>`_.
        :returns: An instance of our model object.
        """
        try:
            last_pk = queryset.order_by('-pk')[0].pk
        except IndexError:
            return

        queryset = queryset.order_by('pk')
        pk = queryset[0].pk - 1
        while pk < last_pk:
            for row in queryset.filter(pk__gt=pk)[:chunksize]:
                pk = row.pk
                yield row
            gc.collect()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
from setuptools import setup, find_packages


__version__ = None
with open('src/printen/version.py') as vfp:
    vd = vfp.read().strip()
    __version__ = vd.split('=')[1].strip().strip('\'').strip('"')


readme = open('README.rst').read()
history = open('HISTORY.rst').read()
long_description = readme + '\n\n' + history


def get_requirements(filename):
    requirements = []
    if os.path.exists(filename):
        with open(filename) as rfp:
            [
                requirements.append(r.strip())
                for r in rfp if not r.startswith('-')
            ]

    return requirements


requirements = get_requirements('requirements.txt')
test_requirements = get_requirements('requirements_test.txt')


setup(
    name='printen',
    version=__version__ if __version__ else 'UNKNOWN',
    description='Flexible integration of Elasticsearch into Python projects.',
    long_description=long_description,
    author='James Kelly',
    author_email='pthread1981@gmail.com',
    url='https://github.com/jimjkelly/printen',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    install_requires=requirements,
    license='BSD',
    zip_safe=False,
    keywords='elasticsearch search indexing flask django peewee',
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: BSD License',
        'Intended Audience :: Developers',
        'Environment :: Web Environment',
        'Programming Language :: Python',
        'Framework :: Flask',
        'Framework :: Django',
        'Topic :: Internet :: WWW/HTTP :: WSGI',
        'Natural Language :: English',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
    ],
    tests_require=test_requirements,
    test_suite='nose.collector'
)



from unittest import TestCase
from elasticsearch import Elasticsearch
from printen.exceptions import MissingObjectError
from printen.mixins import (
    ElasticsearchIndexMixin,
    DjangoElasticsearchIndexMixin,
    PeeweeElasticsearchIndexMixin
)
import printen.config


ELASTICSEARCH_SETTINGS = {
    'ELASTICSEARCH_DELETE_OLD_INDICES': True,
    'ELASTICSEARCH_INDEX_NAME': 'test',
    'ELASTICSEARCH_TYPE_CLASSES': [],
    'ELASTICSEARCH_CONNECTION_PARAMS': {
        'hosts': 'http://127.0.0.1:9200',
    },
}


class MixinTests(TestCase):
    def tearDown(self):
        # Ensure we have a clean settings object
        # between runs
        printen.config.settings = printen.config.SettingsObject()

    def test_get_es(self):
        # An unconfigured model should not have
        # an es object.
        with self.assertRaises(AttributeError):
            ElasticsearchIndexMixin.get_es()

        printen.config.configure(ELASTICSEARCH_SETTINGS)

        self.assertTrue(isinstance(
            ElasticsearchIndexMixin.get_es(),
            Elasticsearch
        ))

    def test_get_es_connection_settings(self):
        with self.assertRaises(AttributeError):
            ElasticsearchIndexMixin.get_es_connection_settings()

        printen.config.configure(ELASTICSEARCH_SETTINGS)

        self.assertEqual(
            ElasticsearchIndexMixin.get_es_connection_settings(),
            ELASTICSEARCH_SETTINGS['ELASTICSEARCH_CONNECTION_PARAMS']
        )

    def test_get_index_name(self):
        self.assertEqual(
            ElasticsearchIndexMixin.get_index_name(),
            'index'
        )

        printen.config.configure(ELASTICSEARCH_SETTINGS)

        self.assertEqual(
            ElasticsearchIndexMixin.get_index_name(),
            ELASTICSEARCH_SETTINGS['ELASTICSEARCH_INDEX_NAME']
        )

    def test_get_document_id(self):
        # The base class doesn't implement this at all
        with self.assertRaises(NotImplementedError):
            ElasticsearchIndexMixin.get_document_id(object())

        # The Django (and Peewee) versions do - but of course
        # an object without the right attribute will fail.
        with self.assertRaises(MissingObjectError):
            DjangoElasticsearchIndexMixin.get_document_id(object())

        with self.assertRaises(MissingObjectError):
            PeeweeElasticsearchIndexMixin.get_document_id(object())

        # Now we create objects we expect to succeed:
        class Foo(object):
            pass
        foo = Foo()
        foo.pk = 1
        self.assertEqual(
            DjangoElasticsearchIndexMixin.get_document_id(foo),
            1
        )

        class Bar(object):
            pass
        bar = Bar()
        bar._meta = Bar()
        bar._meta.primary_key = Bar()
        bar._meta.primary_key.name = 'baz'
        bar.baz = 1
        self.assertEqual(
            PeeweeElasticsearchIndexMixin.get_document_id(bar),
            1
        )

    def test_get_type_name(self):
        self.assertEqual(
            ElasticsearchIndexMixin.get_type_name(),
            'ElasticsearchIndexMixin'
        )

        class Object(object):
            pass
        DjangoElasticsearchIndexMixin._meta = Object()
        DjangoElasticsearchIndexMixin._meta.app_label = 'foo'
        DjangoElasticsearchIndexMixin._meta.model_name = 'bar'
        self.assertEqual(
            DjangoElasticsearchIndexMixin.get_type_name(),
            'foo.bar'
        )

        PeeweeElasticsearchIndexMixin.model = Object()
        PeeweeElasticsearchIndexMixin.model._meta = Object()
        PeeweeElasticsearchIndexMixin.model._meta.name = 'foo'
        self.assertEqual(
            PeeweeElasticsearchIndexMixin.get_type_name(),
            'foo'
        )



from mock import patch
from unittest import TestCase
from printen.mixins import ElasticsearchIndexMixin
from printen.exceptions import Misconfigured
import printen.config
import printen.mixins
import printen.utils


ELASTICSEARCH_SETTINGS = {
    'ELASTICSEARCH_DELETE_OLD_INDICES': True,
    'ELASTICSEARCH_INDEX_NAME': 'test',
    'ELASTICSEARCH_TYPE_CLASSES': [],
    'ELASTICSEARCH_CONNECTION_PARAMS': {
        'hosts': 'http://127.0.0.1:9200',
    },
}


class DummyIndex(ElasticsearchIndexMixin):
    pass


class UtilsTests(TestCase):
    def setUp(self):
        printen.config.configure(ELASTICSEARCH_SETTINGS)

    def tearDown(self):
        # Ensure we have a clean settings object
        # between runs
        printen.config.settings = printen.config.SettingsObject()

    def test_get_indices(self):
        # This will raise a misconfigured when nothing
        # is configured for type classes yet.
        with self.assertRaises(Misconfigured):
            printen.utils.get_indices()

        printen.config.configure({
            'ELASTICSEARCH_TYPE_CLASSES': [
                'some.missing.class'
            ]
        })

        # And it will still raise if they're wrong
        with self.assertRaises(Misconfigured):
            printen.utils.get_indices()

        printen.config.configure({
            'ELASTICSEARCH_TYPE_CLASSES': [
                '{}.DummyIndex'.format(__name__)
            ]
        })

        self.assertEqual(printen.utils.get_indices(), {
            'index': [DummyIndex]
        })

    @patch('printen.utils.es.get_aliases', autospec=True)
    def test_get_indices_from_aliases(self, mock_get_aliases):
        get_aliases_return = {
            'test-123455': {
                'aliases': {
                    'test': {}
                }
            }
        }
        mock_get_aliases.return_value = get_aliases_return

        self.assertEqual(
            printen.utils.get_indices_from_aliases(search_aliases=['test']),
            ['test-123455']
        )

    def test_get_alias_names(self):
        self.assertEqual(
            printen.utils.get_alias_names([('test', 'test-123455')]),
            ['test']
        )

        self.assertEqual(
            printen.utils.get_alias_names([]),
            []
        )

        self.assertEqual(
            printen.utils.get_alias_names([
                ('test', 'test-123456'),
                ('test', 'test-234567')
            ]),
            ['test']
        )

        self.assertEqual(
            printen.utils.get_alias_names([
                ('test', 'test-123456'),
                ('tests', 'test-234567')
            ]),
            ['test', 'tests']
        )

    @patch('printen.utils.es.get_aliases', autospec=True)
    @patch('printen.utils.es.update_aliases', autospec=True)
    def test_create_aliases(self, mock_update_aliases, mock_get_aliases):
        get_aliases_return = {
            'test-123456': {
                'aliases': {
                    'test': {}
                }
            }
        }
        actions = {
            'actions': [
                {
                    'remove': {
                        'index': 'test-123456',
                        'alias': 'test'
                    }
                },
                {
                    'add': {
                        'index': 'test-234567',
                        'alias': 'test'
                    }
                }
            ]
        }

        mock_get_aliases.return_value = get_aliases_return
        printen.utils.create_aliases([('test', 'test-234567')])
        mock_update_aliases.assert_called_with(actions)

    def test_find_search_types(self):
        # A fairly meh test, but meh.
        self.assertItemsEqual(
            printen.utils.find_search_types(printen),
            [
                printen.mixins.PeeweeElasticsearchIndexMixin,
                printen.mixins.DjangoElasticsearchIndexMixin
            ]
        )
